/*! *********************************************************************************
 * \addtogroup Wireless UART Application
 * @{
 ********************************************************************************** */
/*! *********************************************************************************
* Copyright (c) 2015, Freescale Semiconductor, Inc.
* Copyright 2016-2017 NXP
* All rights reserved.
*
* \file
*
* This file is the source file for the Wireless UART Application
*
* SPDX-License-Identifier: BSD-3-Clause
********************************************************************************** */

/************************************************************************************
 *************************************************************************************
 * Include
 *************************************************************************************
 ************************************************************************************/
/* Framework / Drivers */
#include "RNG_Interface.h"
#include "Keyboard.h"
#include "LED.h"
#include "TimersManager.h"
#include "FunctionLib.h"
#include "Panic.h"
#include "SerialManager.h"
#include "MemManager.h"
#include "board.h"

/* BLE Host Stack */
#include "gatt_interface.h"
#include "gatt_server_interface.h"
#include "gatt_client_interface.h"
#include "gatt_database.h"
#include "gap_interface.h"
#include "gatt_db_app_interface.h"

#ifndef MULTICORE_HOST
#include "gatt_db_handles.h"
#else
#define UUID128(name, ...)\
    extern uint8_t name[16];
#include "gatt_uuid128.h"
#undef UUID128
#endif

/* Profile / Services */
#include "wireless_uart_interface.h"
#include "battery_interface.h"
/* Wrappers */
#include "ble_conn_manager.h"
#include "ble_service_discovery.h"

#include "board.h"
#include "ApplMain.h"
#include "wireless_uart.h"

#ifdef MULTICORE_HOST
#include "erpc_host.h"
#include "dynamic_gatt_database.h"
#endif

#include "fsl_lin.h"

#include "lin_cfg.h"

/************************************************************************************
 *************************************************************************************
 * Private macros
 *************************************************************************************
 ************************************************************************************/

#define mAppUartBufferSize_c   			gAttMaxWriteDataSize_d(gAttMaxMtu_c) /* Local Buffer Size */

#define mAppUartFlushIntervalInMs_c   	(7) 	/* Flush Timeout in Ms */

#define mBatteryLevelReportInterval_c   (10)	/* battery level report interval in seconds  */
   
#define MASTER_INSTANCE_LIN LI0_Master
#define LI0_Master 0x01

/************************************************************************************
 *************************************************************************************
 * Private type definitions
 *************************************************************************************
 ************************************************************************************/
typedef enum appEvent_tag{
    mAppEvt_PeerConnected_c,
    mAppEvt_PairingComplete_c,
    mAppEvt_ServiceDiscoveryComplete_c,
    mAppEvt_ServiceDiscoveryNotFound_c,
    mAppEvt_ServiceDiscoveryFailed_c,
    mAppEvt_GattProcComplete_c,
    mAppEvt_GattProcError_c
}appEvent_t;

typedef enum appState_tag
{
    mAppIdle_c, 
    mAppExchangeMtu_c,	
    mAppServiceDisc_c,
    mAppServiceDiscRetry_c,
    mAppRunning_c
} appState_t;

typedef struct appPeerInfo_tag
{
    deviceId_t  deviceId;
    bool_t      isBonded;
    wucConfig_t clientInfo;
    appState_t  appState;
    gapRole_t   gapRole;
} appPeerInfo_t;

typedef struct advState_tag
{
    bool_t advOn;
} advState_t;
/************************************************************************************
 *************************************************************************************
 * Private memory declarations
 *************************************************************************************
 ************************************************************************************/

static appPeerInfo_t maPeerInformation[gAppMaxConnections_c];
static gapRole_t     mGapRole;

/* Adv Parmeters */
static advState_t mAdvState;
static bool_t   mScanningOn = FALSE;

static uint16_t mCharMonitoredHandles[1] = { value_uart_stream };

/* Service Data*/
static wusConfig_t mWuServiceConfig;
static bool_t      mBasValidClientList[gAppMaxConnections_c] = {FALSE};
static basConfig_t mBasServiceConfig = {service_battery, 0, mBasValidClientList, gAppMaxConnections_c};

static tmrTimerID_t mAppTimerId = gTmrInvalidTimerID_c;
static tmrTimerID_t mBatteryMeasurementTimerId = gTmrInvalidTimerID_c;

static tmrTimerID_t mErrorCommandTimerID = gTmrInvalidTimerID_c;

static uint16_t mAppUartBufferSize = mAppUartBufferSize_c;
static volatile bool_t mAppUartNewLine = FALSE;

uint8_t linprotocol_string[3] = {'l', 'i', 'n'};
uint8_t canprotocol_string[3] = {'c', 'a', 'n'};

uint8_t help_string[4] = {'h', 'e', 'l', 'p'};

uint8_t linon_string[2] = {'o', 'n'};
uint8_t linoff_string[3] = {'o', 'f', 'f'};

uint8_t success_string[9] = {'s', 'u', 'c', 'c', 'e', 's', 's', '\r', '\n'};
uint8_t success_string_size=9;

uint8_t command_list_string_1[] = {'A', 'v', 'a', 'i', 'l', 'a','b','l','e',' ','c','o','m','m','a','n','d','s',':','\r', '\n', '\r', '\n',
                                   'l', 'i', 'n', ' ', 'o', 'n', '\r', '\n',
                                   'l', 'i', 'n', ' ', 'o', 'f', 'f', '\r', '\n', '\r', '\n'};

uint8_t command_list_string_size_1 = 42;

uint8_t command_list_string_2[] = {  'l', 'i', 'n', ' ', 't', 'o','g','g','l','e',' ','l','e','d',' ','r','e','d', '\r', '\n',
                                     'l', 'i', 'n', ' ', 't', 'o','g','g','l','e',' ','l','e','d',' ','g','r','e','e','n', '\r', '\n',
                                     'l', 'i', 'n', ' ', 't', 'o','g','g','l','e',' ','l','e','d',' ','b','l','u','e', '\r', '\n',
                                     'l', 'i', 'n', ' ', 'g', 'e','t',' ','l','e','d', '\r', '\n',
                                     'l', 'i', 'n', ' ', 'g', 'e','t',' ','t','e','m','p','e','r','a','t','u','r','e', '\r', '\n','\r', '\n' };

uint8_t command_list_string_size_2 = 99;


uint8_t command_list_string_3[] = {  'c', 'a', 'n', ' ', 't', 'o','g','g','l','e',' ','l','e','d',' ','r','e','d', '\r', '\n',
                                     'c', 'a', 'n', ' ', 't', 'o','g','g','l','e',' ','l','e','d',' ','g','r','e','e','n', '\r', '\n',
                                     'c', 'a', 'n', ' ', 't', 'o','g','g','l','e',' ','l','e','d',' ','b','l','u','e', '\r', '\n',
                                     'c', 'a', 'n', ' ', 'g', 'e','t',' ','l','e','d', '\r', '\n',
                                     'c', 'a', 'n', ' ', 'g', 'e','t',' ','t','e','m','p','e','r','a','t','u','r','e', '\r', '\n','\r', '\n',
                                     'n','o','t','e',':',' ', 'c', 'a', 's', 'e', ' ','s','e','n','s','i','t','i','v','e','\r', '\n'};

uint8_t command_list_string_size_3 = 121;

uint8_t sleeping_string[17] = {'l', 'i', 'n', ' ', 'i', 's', ' ','s','l','e','e','p','i','n','g', '\r', '\n'};
uint8_t sleeping_string_size = 17;

uint8_t error_string[15] = {'e', 'r', 'r', 'o', 'r', ' ', 'c', 'o', 'm', 'm', 'a', 'n', 'd', '\r', '\n'};
uint8_t error_string_size=15;

uint8_t ledred_string[14] = {'t', 'o', 'g', 'g', 'l', 'e', ' ', 'l', 'e', 'd', ' ', 'r', 'e', 'd'};
uint8_t ledgreen_string[16] = {'t', 'o', 'g', 'g', 'l', 'e', ' ', 'l', 'e', 'd', ' ', 'g', 'r', 'e', 'e', 'n'};
uint8_t ledblue_string[15] = {'t', 'o', 'g', 'g', 'l', 'e', ' ', 'l', 'e', 'd', ' ', 'b', 'l', 'u', 'e'};
uint8_t gettemp_string[15] = {'g', 'e', 't', ' ', 't', 'e', 'm', 'p', 'e', 'r', 'a', 't', 'u', 'r', 'e'};
uint8_t getled_string[7] = {'g', 'e', 't', ' ', 'l', 'e', 'd'};

uint8_t sw2_can_string[13] = {'s', 'w', '2', ' ', 'v', 'i', 'a', ' ', 'c', 'a', 'n', '\r', '\n'};
uint8_t sw2_can_string_size=13;

uint8_t sw3_can_string[13] = {'s', 'w', '3', ' ', 'v', 'i', 'a', ' ', 'c', 'a', 'n', '\r', '\n'};
uint8_t sw3_can_string_size=13;

uint8_t sw2_lin_string[13] = {'s', 'w', '2', ' ', 'v', 'i', 'a', ' ', 'l', 'i', 'n', '\r', '\n'};
uint8_t sw2_lin_string_size=13;

uint8_t sw3_lin_string[13] = {'s', 'w', '3', ' ', 'v', 'i', 'a', ' ', 'l', 'i', 'n', '\r', '\n'};
uint8_t sw3_lin_string_size=13;

uint8_t temperature_string[25] = {'t', 'e', 'm', 'p', 'e', 'r', 'a', 't', 'u', 'r', 'e', ':', ' ', '0', '0', ' ', 'c', 'e', 'l','s','i','u','s', '\r', '\n'};
uint8_t temperature_string_size = 25;

uint8_t ledredstatus_string[19] = {'l', 'e', 'd', ' ', 'r', 'e', 'd', ' ', 's', 't', 'a', 't', 'u', 's', ':', ' ', ' ', '\r', '\n'};
uint8_t ledred_status_string_size = 19;

uint8_t ledgreenstatus_string[21] = {'l', 'e', 'd', ' ', 'g', 'r', 'e', 'e', 'n', ' ', 's', 't', 'a', 't', 'u', 's', ':', ' ', ' ', '\r', '\n'};
uint8_t ledgreen_status_string_size = 21;

uint8_t ledbluestatus_string [20] = {'l', 'e', 'd', ' ', 'b', 'l', 'u', 'e', ' ', 's', 't', 'a', 't', 'u', 's', ':', ' ', ' ', '\r', '\n'};
uint8_t ledblue_status_string_size = 20;

uint8_t seq_number_ledred, seq_number_ledgreen, seq_number_ledblue, seq_number_getled, seq_number_gettemp = 0;

uint8_t seq_number_sendled_new, seq_number_sendled_last = 0;
uint8_t seq_number_sendtemp_new, seq_number_sendtemp_last = 0;
uint8_t seq_number_sw2_new, seq_number_sw2_last = 0;
uint8_t seq_number_sw3_new, seq_number_sw3_last = 0;
uint8_t seq_number_success_new, seq_number_success_last = 0;

extern uint8_t first_time_command_flag;

uint8_t errorcommand_timeout_flag;

uint8_t ones_temp;
uint8_t tens_temp;

/************************************************************************************
 *************************************************************************************
 * Private functions prototypes
 *************************************************************************************
 ************************************************************************************/

/* Gatt and Att callbacks */
static void BleApp_AdvertisingCallback(gapAdvertisingEvent_t* pAdvertisingEvent);
static void BleApp_ScanningCallback (gapScanningEvent_t* pScanningEvent);
static void BleApp_ConnectionCallback
(
    deviceId_t peerDeviceId,
    gapConnectionEvent_t* pConnectionEvent
);
static void BleApp_GattServerCallback 
(
   deviceId_t deviceId,
   gattServerEvent_t* pServerEvent
);

static void BleApp_GattClientCallback 
(
    deviceId_t              serverDeviceId,
    gattProcedureType_t     procedureType,
    gattProcedureResult_t   procedureResult,
    bleResult_t             error
);

static void BleApp_ServiceDiscoveryCallback
(
    deviceId_t deviceId,
    servDiscEvent_t* pEvent
);

static void BleApp_Config (void);
static void BleApp_Advertise (void);

void BleApp_StateMachineHandler
(
    deviceId_t peerDeviceId,
    uint8_t event
);

static void BleApp_StoreServiceHandles
(
    deviceId_t 	     peerDeviceId,
    gattService_t   *pService
);
static bool_t BleApp_CheckScanEvent (gapScannedDevice_t* pData);

/* Timer Callbacks */
static void ScaningTimerCallback (void *);
static void BatteryMeasurementTimerCallback (void *);

static void BleApp_ReceivedUartStream(deviceId_t peerDeviceId, uint8_t *pStream, uint16_t streamLength);

void APP_ProcessString(uint8_t *command_string);

static int32_t StrCompare(const char *str1, const char *str2, int32_t count); /*!< compare string command */

extern void lin_go_to_sleep(void);

extern uint8_t check_if_lin_sleeping(void);

extern void app_tx_can(uint8_t databyte0, uint8_t databyte1, uint8_t databyte2, uint8_t databyte3);

static void mErrorCommandTimerCb(uint8_t timerId);

/************************************************************************************
*************************************************************************************
* Public functions
*************************************************************************************
************************************************************************************/

/*! *********************************************************************************
* \brief    Initializes application specific functionality before the BLE stack init.
*
********************************************************************************** */
void BleApp_Init(void)
{
    /* Initialize application support for drivers */
    BOARD_InitAdc();
    
#ifdef MULTICORE_HOST
    /* Init eRPC host */
    init_erpc_host();
#endif    
}

/*! *********************************************************************************
 * \brief    Starts the BLE application.
 *
 * \param[in]    mGapRole    GAP Start Role (Central or Peripheral).
 ********************************************************************************** */
void BleApp_Start (gapRole_t mGapRole)
{
    switch (mGapRole)
    {
        case gGapCentral_c:
        {
            mAppUartNewLine = TRUE;
            gPairingParameters.localIoCapabilities = gIoKeyboardDisplay_c;
            App_StartScanning(&gScanParams, BleApp_ScanningCallback, TRUE);
            break;
        }
        case gGapPeripheral_c:
        {
            mAppUartNewLine = TRUE;
            gPairingParameters.localIoCapabilities = gIoDisplayOnly_c;
            BleApp_Advertise();
            break;
        }
        default:
            break;
    }
}

/*! *********************************************************************************
* \brief        Handles keyboard events.
*
* \param[in]    events    Key event structure.
********************************************************************************** */
void BleApp_HandleKeys(key_event_t events)
{
    uint8_t mPeerId = 0;
    switch (events)
    {
        case gKBD_EventPressPB1_c:
        {
            LED_StopFlashingAllLeds();
            Led1Flashing();

            BleApp_Start(mGapRole);
            break;
        }
        case gKBD_EventLongPB1_c:
        {
            for(mPeerId = 0; mPeerId < gAppMaxConnections_c; mPeerId++)
            {
                if (maPeerInformation[mPeerId].deviceId != gInvalidDeviceId_c)
                    Gap_Disconnect(maPeerInformation[mPeerId].deviceId);
            }
            break;
        }
        case gKBD_EventPressPB2_c:
            break;
        case gKBD_EventLongPB2_c:
            break;
    default:
        break;
    }
}

/*! *********************************************************************************
* \brief        Handles BLE generic callback.
*
* \param[in]    pGenericEvent    Pointer to gapGenericEvent_t.
********************************************************************************** */
void BleApp_GenericCallback (gapGenericEvent_t* pGenericEvent)
{
    /* Call BLE Conn Manager */
    BleConnManager_GenericEvent(pGenericEvent);
    
    switch (pGenericEvent->eventType)
    {
        case gInitializationComplete_c:    
        {
            BleApp_Config();
        }
        break;    
        
        case gAdvertisingParametersSetupComplete_c:
        {
            App_StartAdvertising(BleApp_AdvertisingCallback, BleApp_ConnectionCallback);
        }
        break;    
        
        default: 
            break;
    }
}

/************************************************************************************
*************************************************************************************
* Private functions
*************************************************************************************
************************************************************************************/

/*! *********************************************************************************
* \brief        Configures BLE Stack after initialization. Usually used for
*               configuring advertising, scanning, white list, services, et al.
*
********************************************************************************** */
static void BleApp_Config()
{
    uint8_t mPeerId = 0;
    
#ifdef MULTICORE_HOST
    if (GattDbDynamic_CreateDatabase() != gBleSuccess_c)
    {
        panic(0,0,0,0);
        return;
    }
#endif /* MULTICORE_HOST */

    /* Configure as GAP Dual Role */
    BleConnManager_GapDualRoleConfig();
        
    /* Register for callbacks */
    App_RegisterGattServerCallback(BleApp_GattServerCallback);
    App_RegisterGattClientProcedureCallback(BleApp_GattClientCallback);
    GattServer_RegisterHandlesForWriteNotifications(NumberOfElements(mCharMonitoredHandles), mCharMonitoredHandles);
    BleServDisc_RegisterCallback(BleApp_ServiceDiscoveryCallback);
	
    for(mPeerId = 0; mPeerId < gAppMaxConnections_c; mPeerId++)
    {
        maPeerInformation[mPeerId].appState = mAppIdle_c;
        maPeerInformation[mPeerId].deviceId = gInvalidDeviceId_c;
        maPeerInformation[mPeerId].clientInfo.hService = gGattDbInvalidHandleIndex_d;
        maPeerInformation[mPeerId].clientInfo.hUartStream = gGattDbInvalidHandleIndex_d;
    }
    
    /* start node as GAP peripheral */
	mGapRole = gGapPeripheral_c;
    
    
    mAdvState.advOn = FALSE;
    mScanningOn = FALSE;
    
    /* Start services */
    Wus_Start(&mWuServiceConfig);

    mBasServiceConfig.batteryLevel = BOARD_GetBatteryLevel();
    Bas_Start(&mBasServiceConfig);

    /* Allocate application timer */
    mAppTimerId = TMR_AllocateTimer();
    
    mErrorCommandTimerID = TMR_AllocateTimer(); //allocate timer for error command

}

/*! *********************************************************************************
* \brief        Configures GAP Advertise parameters. Advertise will satrt after
*               the parameters are set.
*
********************************************************************************** */
static void BleApp_Advertise(void)
{
    /* Set advertising parameters*/
    Gap_SetAdvertisingParameters(&gAdvParams);
}

/*! *********************************************************************************
 * \brief        Handles BLE Scanning callback from host stack.
 *
 * \param[in]    pScanningEvent    Pointer to gapScanningEvent_t.
 ********************************************************************************** */
static void BleApp_ScanningCallback (gapScanningEvent_t* pScanningEvent)
{
    switch (pScanningEvent->eventType)
    {
        case gDeviceScanned_c:
        {
            if (BleApp_CheckScanEvent(&pScanningEvent->eventData.scannedDevice))
            {        
                gConnReqParams.peerAddressType = pScanningEvent->eventData.scannedDevice.addressType;
                FLib_MemCpy(gConnReqParams.peerAddress, 
                            pScanningEvent->eventData.scannedDevice.aAddress,
                            sizeof(bleDeviceAddress_t));
                
                Gap_StopScanning();
#if gAppUsePrivacy_d
                gConnReqParams.usePeerIdentityAddress = pScanningEvent->eventData.scannedDevice.advertisingAddressResolved;
#endif
                App_Connect(&gConnReqParams, BleApp_ConnectionCallback);
            }
        }        
        break;
        
        case gScanStateChanged_c:
        {
            mScanningOn = !mScanningOn;
            
            /* Node starts scanning */
            if (mScanningOn)
            { 
                /* Start advertising timer */
                TMR_StartLowPowerTimer(mAppTimerId, 
                           gTmrLowPowerSecondTimer_c,
                           TmrSeconds(gScanningTime_c),
                           ScaningTimerCallback, NULL);  

                Led1Flashing();
            }
            /* Node is not scanning */
            else
            {                
                TMR_StopTimer(mAppTimerId);
                
                Led1Flashing();
                Led2Flashing();
                Led3Flashing();
                Led4Flashing();
            }
        }
        break;
    case gScanCommandFailed_c:
    {
        panic(0, 0, 0, 0);
        break;
    }
    default:
        break;
    }
}

/*! *********************************************************************************
* \brief        Handles BLE Advertising callback from host stack.
*
* \param[in]    pAdvertisingEvent    Pointer to gapAdvertisingEvent_t.
********************************************************************************** */
static void BleApp_AdvertisingCallback (gapAdvertisingEvent_t* pAdvertisingEvent)
{
    switch (pAdvertisingEvent->eventType)
    {
        case gAdvertisingStateChanged_c:
        {
            mAdvState.advOn = !mAdvState.advOn;
            LED_StopFlashingAllLeds();
            Led1Flashing();

            if(!mAdvState.advOn)
            {
                Led2Flashing();
                Led3Flashing();
                Led4Flashing();
            }
        }
        break;

        case gAdvertisingCommandFailed_c:
        {
            panic(0,0,0,0);
        }
        break;

        default:
            break;
    }
}


static void BleApp_SendUartStream(uint8_t *pRecvStream, uint8_t streamSize)
{
    gattCharacteristic_t characteristic;
    uint8_t              mPeerId = 0;

    /* send UART stream to all peers */
    for(mPeerId = 0; mPeerId < gAppMaxConnections_c; mPeerId++)
    {
        if(gInvalidDeviceId_c != maPeerInformation[mPeerId].deviceId && 
           mAppRunning_c == maPeerInformation[mPeerId].appState)
        {
            characteristic.value.handle = maPeerInformation[mPeerId].clientInfo.hUartStream;
            GattClient_WriteCharacteristicValue(mPeerId, &characteristic,
                                                streamSize, pRecvStream, TRUE,
                                                FALSE, FALSE, NULL);
        }
    }
}

/*! *********************************************************************************
* \brief        Handles BLE Connection callback from host stack.
*
* \param[in]    peerDeviceId        Peer device ID.
* \param[in]    pConnectionEvent    Pointer to gapConnectionEvent_t.
********************************************************************************** */
static void BleApp_ConnectionCallback (deviceId_t peerDeviceId, gapConnectionEvent_t* pConnectionEvent)
{
    switch (pConnectionEvent->eventType)
    {
        case gConnEvtConnected_c:
        {
            if(peerDeviceId < gAppMaxConnections_c)
            {
                maPeerInformation[peerDeviceId].deviceId = peerDeviceId;
            }
            else
            {
                Gap_Disconnect(peerDeviceId);
                break;
            }

            /* Advertising stops when connected */
            mAdvState.advOn = FALSE;

            /* Subscribe client*/
            Wus_Subscribe(peerDeviceId);
            Bas_Subscribe(&mBasServiceConfig, peerDeviceId);

            /* UI */
            LED_StopFlashingAllLeds();
            Led1On();

            /* Stop Advertising Timer*/
            mAdvState.advOn = FALSE;
            TMR_StopTimer(mAppTimerId);
            
            if(!TMR_IsTimerActive(mBatteryMeasurementTimerId))
            {
                /* Start battery measurements */
                TMR_StartLowPowerTimer(mBatteryMeasurementTimerId, gTmrLowPowerIntervalMillisTimer_c,
                                       TmrSeconds(mBatteryLevelReportInterval_c), BatteryMeasurementTimerCallback, NULL);
            }

#if gAppUsePairing_d
#if gAppUseBonding_d
            if( mGapRole == gGapCentral_c )
            {
                Gap_CheckIfBonded(peerDeviceId, &maPeerInformation[peerDeviceId].isBonded);
                
                if ((maPeerInformation[peerDeviceId].isBonded) &&
                    (gBleSuccess_c == Gap_LoadCustomPeerInformation(peerDeviceId,
                        (void*) &maPeerInformation[peerDeviceId].clientInfo, 0, sizeof (wucConfig_t))))
                {
                    /* Restored custom connection information. Encrypt link */
                    Gap_EncryptLink(peerDeviceId);
                }
            }
#endif /* gAppUseBonding_d*/ 
#endif /* gAppUsePairing_d */
            
            mAppUartNewLine = TRUE;
            
            maPeerInformation[peerDeviceId].gapRole = mGapRole;

            /* run the state machine */
            BleApp_StateMachineHandler(peerDeviceId, mAppEvt_PeerConnected_c);
        }
        break;

        case gConnEvtDisconnected_c:
        {

            maPeerInformation[peerDeviceId].appState = mAppIdle_c;
            
            /* Unsubscribe client */
            Wus_Unsubscribe();
            Bas_Unsubscribe(&mBasServiceConfig, peerDeviceId); 
            
            TMR_StopTimer(mBatteryMeasurementTimerId);

            /* Reset Service Discovery to be sure*/
            BleServDisc_Stop(peerDeviceId);

            /* UI */
            LED_TurnOffAllLeds();
            LED_StartFlash(LED_ALL);
 
            /* mark device id as invalid */
            maPeerInformation[peerDeviceId].deviceId = gInvalidDeviceId_c;
            
            /* recalculate minimum of maximum MTU's of all connected devices */
            mAppUartBufferSize                       = mAppUartBufferSize_c;
            
            for(uint8_t mPeerId = 0; mPeerId < gAppMaxConnections_c; mPeerId++)
            {
                if(gInvalidDeviceId_c != maPeerInformation[mPeerId].deviceId)
                {
                    uint16_t tempMtu;
                    
                    Gatt_GetMtu(mPeerId, &tempMtu);
                    tempMtu = gAttMaxWriteDataSize_d(tempMtu);
                    
                    if(tempMtu < mAppUartBufferSize)
                    {
                        mAppUartBufferSize = tempMtu;
                    }
                }
            }
            BleApp_Start(mGapRole);
        }
        break;

#if gAppUsePairing_d		
        case gConnEvtPairingComplete_c:
        {
            if (pConnectionEvent->eventData.pairingCompleteEvent.pairingSuccessful)
            {
                BleApp_StateMachineHandler(peerDeviceId, 
                                           mAppEvt_PairingComplete_c);
            }
        }
        break;
#endif /* gAppUsePairing_d */

    default:
        break;
    }
    
    /* Connection Manager to handle Host Stack interactions */
    if (maPeerInformation[peerDeviceId].gapRole == gGapCentral_c)
    {
        BleConnManager_GapCentralEvent(peerDeviceId, pConnectionEvent);
    }
    else if (maPeerInformation[peerDeviceId].gapRole == gGapPeripheral_c)
    {
        BleConnManager_GapPeripheralEvent(peerDeviceId, pConnectionEvent);
    }
}

static void BleApp_ServiceDiscoveryCallback(deviceId_t peerDeviceId, servDiscEvent_t* pEvent)
{
    switch(pEvent->eventType)
    {
        case gServiceDiscovered_c:
        {
            if (pEvent->eventData.pService->uuidType == gBleUuidType128_c)
            {
                if(FLib_MemCmp((void*)&uuid_service_wireless_uart, (void*)&pEvent->eventData.pService->uuid, sizeof(bleUuid_t)))
                { 
                    BleApp_StoreServiceHandles(peerDeviceId, pEvent->eventData.pService); 
                }
            }
        }
        break;

        case gDiscoveryFinished_c:
        {
            if (pEvent->eventData.success)
            {
                if (gGattDbInvalidHandleIndex_d != maPeerInformation[peerDeviceId].clientInfo.hService)
                {
                    BleApp_StateMachineHandler(peerDeviceId, 
                                               mAppEvt_ServiceDiscoveryComplete_c);

                    BleApp_SendUartStream(&welcome_menu_string, welcome_menu_string_size); //welcome menu ble-can-lin bridge demo // @lb

                }
                else
                {
                    BleApp_StateMachineHandler(peerDeviceId, 
                                               mAppEvt_ServiceDiscoveryNotFound_c);                
                }
            }
            else
            {
                BleApp_StateMachineHandler(peerDeviceId, 
                                           mAppEvt_ServiceDiscoveryFailed_c);
            }
        }
        break;

        default:
        break;
    }
}

/*! *********************************************************************************
* \brief        Handles GATT client callback from host stack.
*
* \param[in]    serverDeviceId      GATT Server device ID.
* \param[in]    procedureType    	Procedure type.
* \param[in]    procedureResult    	Procedure result.
* \param[in]    error    			Callback result.
********************************************************************************** */
static void BleApp_GattClientCallback(
    deviceId_t              serverDeviceId,
    gattProcedureType_t     procedureType,
    gattProcedureResult_t   procedureResult,
    bleResult_t             error
)
{  
    if (procedureResult == gGattProcError_c)
    {    
        BleApp_StateMachineHandler(serverDeviceId, 
                                   mAppEvt_GattProcError_c);
    }
    else if (procedureResult == gGattProcSuccess_c)
    {        
    	BleApp_StateMachineHandler(serverDeviceId, 
                                   mAppEvt_GattProcComplete_c);
    }

    /* Signal Service Discovery Module */
    BleServDisc_SignalGattClientEvent(serverDeviceId, 
                                      procedureType,procedureResult, error);
}

/*! *********************************************************************************
 * \brief        Handles GATT server callback from host stack.
 *
 * \param[in]    deviceId        Client peer device ID.
 * \param[in]    pServerEvent    Pointer to gattServerEvent_t.
 ********************************************************************************** */
static void BleApp_GattServerCallback (
                                       deviceId_t deviceId,
                                       gattServerEvent_t* pServerEvent)
{
    uint16_t tempMtu = 0;
    
    switch (pServerEvent->eventType)
    {
        case gEvtAttributeWrittenWithoutResponse_c:
        {
            if (pServerEvent->eventData.attributeWrittenEvent.handle == value_uart_stream)
            {
            	BleApp_ReceivedUartStream(deviceId, pServerEvent->eventData.attributeWrittenEvent.aValue,
                            pServerEvent->eventData.attributeWrittenEvent.cValueLength);
            }
            break;
        }
        case gEvtMtuChanged_c:
        {
            /* update stream length with minimum of  new MTU */
            Gatt_GetMtu(deviceId, &tempMtu);
            tempMtu = gAttMaxWriteDataSize_d(tempMtu);
            
            mAppUartBufferSize = mAppUartBufferSize <= tempMtu? mAppUartBufferSize : tempMtu;
        }
        break;
    default:
        break;
    }
}


static bool_t MatchDataInAdvElementList (
                                         gapAdStructure_t *pElement,
                                         void *pData,
                                         uint8_t iDataLen)
{
    uint8_t i;

    for (i = 0; i < pElement->length; i += iDataLen)
    {
        if (FLib_MemCmp(pData, &pElement->aData[i], iDataLen))
        {
            return TRUE;
        }
    }
    return FALSE;
}

/*! *********************************************************************************
 * \brief        Checks Scan data for a device to connect.
 *
 * \param[in]    pData    Pointer to gapScannedDevice_t.
 ********************************************************************************** */
static bool_t BleApp_CheckScanEvent (gapScannedDevice_t* pData)
{
    uint8_t index = 0;
    bool_t foundMatch = FALSE;

    while (index < pData->dataLength)
    {
        gapAdStructure_t adElement;

        adElement.length = pData->data[index];
        adElement.adType = (gapAdType_t) pData->data[index + 1];
        adElement.aData = &pData->data[index + 2];

        /* Search for Wireless UART Service */
        if ((adElement.adType == gAdIncomplete128bitServiceList_c)
            || (adElement.adType == gAdComplete128bitServiceList_c))
        {
            foundMatch = MatchDataInAdvElementList(&adElement,
                &uuid_service_wireless_uart, 16);
        }

        /* Move on to the next AD element type */
        index += adElement.length + sizeof(uint8_t);
    }

    return foundMatch;
}

/*! *********************************************************************************
 * \brief        Stores handles used by the application.
 *
 * \param[in]    pService    Pointer to gattService_t.
 ********************************************************************************** */
static void BleApp_StoreServiceHandles (deviceId_t peerDeviceId, gattService_t *pService)
{
    /* Found Wireless UART Service */
    maPeerInformation[peerDeviceId].clientInfo.hService = pService->startHandle;

    if (pService->cNumCharacteristics > 0 &&
        pService->aCharacteristics != NULL)
    {
        /* Found Uart Characteristic */
        maPeerInformation[peerDeviceId].clientInfo.hUartStream =
            pService->aCharacteristics[0].value.handle;
    }
}

void BleApp_StateMachineHandler(deviceId_t peerDeviceId, uint8_t event)
{
    uint16_t tempMtu = 0;
    
    /* invalid client information */
    if(gInvalidDeviceId_c == maPeerInformation[peerDeviceId].deviceId)
    {
        return;
    }
    
    switch (maPeerInformation[peerDeviceId].appState)
    {
        case mAppIdle_c:
        {
            if (event == mAppEvt_PeerConnected_c)
            {
                /* Let the central device initiate the Exchange MTU procedure*/
                if (mGapRole == gGapCentral_c)
                {
                    /* Moving to Exchange MTU State */
                    maPeerInformation[peerDeviceId].appState = mAppExchangeMtu_c;
                    GattClient_ExchangeMtu(peerDeviceId);
                }
                else
                {
                    /* Moving to Service Discovery State*/
                    maPeerInformation[peerDeviceId].appState = mAppServiceDisc_c;

                    /* Start Service Discovery*/
                    BleServDisc_FindService(peerDeviceId, 
                                            gBleUuidType128_c,
                                            (bleUuid_t*) &uuid_service_wireless_uart);                  
                }
            }
        }
        break;

        case mAppExchangeMtu_c:
        {
            if (event == mAppEvt_GattProcComplete_c)
            {
                /* update stream length with minimum of maximum MTU's of connected devices */
                Gatt_GetMtu(peerDeviceId, &tempMtu);
                tempMtu = gAttMaxWriteDataSize_d(tempMtu);
                
                mAppUartBufferSize = mAppUartBufferSize <= tempMtu? mAppUartBufferSize : tempMtu;
                
                /* Moving to Service Discovery State*/
                maPeerInformation[peerDeviceId].appState = mAppServiceDisc_c;

                /* Start Service Discovery*/
                BleServDisc_FindService(peerDeviceId, 
                                        gBleUuidType128_c,
                                        (bleUuid_t*) &uuid_service_wireless_uart);
            }
            else if (event == mAppEvt_GattProcError_c) 
            {
                Gap_Disconnect(peerDeviceId);
            }
        }
        break;

        case mAppServiceDisc_c:
        {
            if (event == mAppEvt_ServiceDiscoveryComplete_c)
            {            	
                /* Moving to Running State*/
                maPeerInformation[peerDeviceId].appState = mAppRunning_c;
#if gAppUseBonding_d
                /* Write data in NVM */
                Gap_SaveCustomPeerInformation(maPeerInformation[peerDeviceId].deviceId,
                                              (void*) &maPeerInformation[peerDeviceId].clientInfo, 0,
                                              sizeof (wucConfig_t));
#endif
            }
            else if (event == mAppEvt_ServiceDiscoveryNotFound_c)
            {
                /* Moving to Service discovery Retry State*/
                maPeerInformation[peerDeviceId].appState = mAppServiceDiscRetry_c;              
                /* Retart Service Discovery for all services */
                BleServDisc_Start(peerDeviceId);              
            }
            else if (event == mAppEvt_ServiceDiscoveryFailed_c)
            {
                Gap_Disconnect(peerDeviceId);
            }
        }
        break;
        
        case mAppServiceDiscRetry_c:
        {
            if (event == mAppEvt_ServiceDiscoveryComplete_c)
            {            	
                /* Moving to Running State*/
                maPeerInformation[peerDeviceId].appState = mAppRunning_c;
            }
            else if( (event == mAppEvt_ServiceDiscoveryNotFound_c) ||
                     (event == mAppEvt_ServiceDiscoveryFailed_c) )
            {
                Gap_Disconnect(peerDeviceId);
            }
        }
        break;        

        case mAppRunning_c:
                break;
        default:
                break;
    }
}

/*! *********************************************************************************
 * \brief        Handles scanning timer callback.
 *
 * \param[in]    pParam        Calback parameters.
 ********************************************************************************** */
static void ScaningTimerCallback (void * pParam)
{
    /* Stop scanning */
    Gap_StopScanning();
}


static void BleApp_ReceivedUartStream(deviceId_t peerDeviceId, uint8_t *pStream, uint16_t streamLength)
{
    static deviceId_t previousDeviceId = gInvalidDeviceId_c;
    
    uint8_t  additionalInfoBuff[10] = { '\r','\n','[', '0', '0', '-', 'M', ']', ':', ' '};
    uint8_t *pBuffer = NULL;
    uint8_t  messageHeaderSize = 0;

    if(gInvalidDeviceId_c == peerDeviceId)
    {
        return;
    }
    
    if( (previousDeviceId != peerDeviceId) || mAppUartNewLine )
    {
        streamLength += sizeof(additionalInfoBuff);
    }
    
    /* Allocate buffer for asynchronous write */
    pBuffer = MEM_BufferAlloc(streamLength);

    if (pBuffer != NULL)
    {
        /* if this is a message from a previous device, print device ID */
        if( (previousDeviceId != peerDeviceId) || mAppUartNewLine )
        {
            messageHeaderSize = sizeof(additionalInfoBuff);
            
            if( mAppUartNewLine )
            {
                mAppUartNewLine = FALSE;
            }
            
            if(peerDeviceId > 9)
            {
                additionalInfoBuff[3] = '0' + (peerDeviceId / 10);
            }
            additionalInfoBuff[4] = '0' + (peerDeviceId % 10);
            
            if( gGapCentral_c != maPeerInformation[peerDeviceId].gapRole )
            {
                additionalInfoBuff[6] = 'S';
            }
            
            FLib_MemCpy(pBuffer, additionalInfoBuff, sizeof(additionalInfoBuff));
        }
        
        FLib_MemCpy(pBuffer + messageHeaderSize, pStream, streamLength - messageHeaderSize);

        APP_ProcessString(pStream);

        //Serial_AsyncWrite(gAppSerMgrIf, pBuffer, streamLength, Uart_TxCallBack, pBuffer);
    }
    /* update the previous device ID */
    previousDeviceId = peerDeviceId;
}

/*! *********************************************************************************
* \brief        Handles battery measurement timer callback.
*
* \param[in]    pParam        Calback parameters.
********************************************************************************** */
static void BatteryMeasurementTimerCallback(void * pParam)
{
    mBasServiceConfig.batteryLevel = BOARD_GetBatteryLevel();
    Bas_RecordBatteryMeasurement(&mBasServiceConfig);
}

static int32_t StrCompare(const char *str1, const char *str2, int32_t count)
{
    while (count--)
    {
        if (*str1++ != *str2++)
        {
            return *(unsigned char *)(str1 - 1) - *(unsigned char *)(str2 - 1);
        }
    }
    return 0;
}

void APP_ProcessCommand_Can(uint8_t command, uint8_t byte1, uint8_t byte2, uint8_t byte3)
{

  switch (command){
    case gCommand_SW2_c:
      BleApp_SendUartStream(&sw2_can_string, sw2_can_string_size);

    break;
    case gCommand_SW3_c:
      BleApp_SendUartStream(&sw3_can_string, sw3_can_string_size);

    break;
    case gCommand_SendLedStatus_c:

          ledredstatus_string[16]=byte1+0x30;
          ledgreenstatus_string[18]=byte2+0x30;
          ledbluestatus_string[17]=byte3+0x30;

          BleApp_SendUartStream(&ledredstatus_string, ledred_status_string_size);
          BleApp_SendUartStream(&ledgreenstatus_string, ledgreen_status_string_size);
          BleApp_SendUartStream(&ledbluestatus_string, ledblue_status_string_size);

    break;
    case gCommand_SendTemperature_c:

         ones_temp = byte1%10;
         byte1 = byte1/10;
         tens_temp = byte1%10;

         temperature_string[14] = ones_temp+0x30;
         temperature_string[13] = tens_temp+0x30;

         BleApp_SendUartStream(&temperature_string, temperature_string_size);

    break;
    case gCommand_SendSucessStatus_c:
             
          BleApp_SendUartStream(&success_string, success_string_size);          
      
    break;    
    
    default:
    break;        
  }
}

void APP_ProcessCommand_Lin(uint8_t FrameID) 
{

  switch (FrameID){
    case gID_SW2_SW3_c: 

      if (g_lin_frame_data_buffer[4] == gCommand_SW2_c)
      {
              
              seq_number_sw2_new = g_lin_frame_data_buffer[5];
              
              if (seq_number_sw2_new != seq_number_sw2_last)
              {
                  
                  seq_number_sw2_last = g_lin_frame_data_buffer[5];
                  
                  BleApp_SendUartStream(&sw2_lin_string, sw2_lin_string_size);
                  
              }
      
      } else if (g_lin_frame_data_buffer[4] == gCommand_SW3_c)
      {
              seq_number_sw3_new = g_lin_frame_data_buffer[5];
              
              if (seq_number_sw3_new != seq_number_sw3_last)
              {
                  
                  seq_number_sw3_last = g_lin_frame_data_buffer[5];
                  
                  BleApp_SendUartStream(&sw3_lin_string, sw3_lin_string_size);
                  
              }
      }

    break;
    case gID_SendTemperature_c:   

      if (g_lin_frame_data_buffer[6] == gCommand_SendTemperature_c)
      {
         
        seq_number_sendtemp_new = g_lin_frame_data_buffer[7];
            
        if (seq_number_sendtemp_new != seq_number_sendtemp_last)
        {
          
             seq_number_sendtemp_last = g_lin_frame_data_buffer[7];
             
             ones_temp = g_lin_frame_data_buffer[8]%10;
             g_lin_frame_data_buffer[8] = g_lin_frame_data_buffer[8]/10;
             tens_temp = g_lin_frame_data_buffer[8]%10;
                
             temperature_string[14] = ones_temp+0x30;
             temperature_string[13] = tens_temp+0x30;
             
             BleApp_SendUartStream(&temperature_string, temperature_string_size);
         
        }  
        
      }
      
    break; 
    
    case gID_SendLedStatus_c: 
      
      if (g_lin_frame_data_buffer[9] == gCommand_SendLedStatus_c)
      {
        
            seq_number_sendled_new = g_lin_frame_data_buffer[10];
            
              if (seq_number_sendled_new != seq_number_sendled_last)
              {
                  
                  seq_number_sendled_last = g_lin_frame_data_buffer[10];

                  ledredstatus_string[16]=g_lin_frame_data_buffer[11]+0x30;
                  ledgreenstatus_string[18]=g_lin_frame_data_buffer[12]+0x30;
                  ledbluestatus_string[17]=g_lin_frame_data_buffer[13]+0x30;
                  
                  BleApp_SendUartStream(&ledredstatus_string, ledred_status_string_size);
                  BleApp_SendUartStream(&ledgreenstatus_string, ledgreen_status_string_size);
                  BleApp_SendUartStream(&ledbluestatus_string, ledblue_status_string_size);

             }
                
      }
      
     break;
     
    case gID_SendSucessStatus_c:
      
      if (g_lin_frame_data_buffer[14] == gCommand_SendSucessStatus_c)
      {
          
          seq_number_success_new = g_lin_frame_data_buffer[15];
            
          if (seq_number_success_new != seq_number_success_last)
          {
              
              seq_number_success_last = g_lin_frame_data_buffer[15];
              
              BleApp_SendUartStream(&success_string, success_string_size);   
          }
      }

    break;

    default:
    break;
  }
}


void APP_ProcessString(uint8_t *command_string){

  if (!errorcommand_timeout_flag)
  {

  /*check if "help" was typed*/
  if (StrCompare(command_string, &help_string[0], 4) == 0){

      BleApp_SendUartStream(&command_list_string_1, command_list_string_size_1);
      BleApp_SendUartStream(&command_list_string_2, command_list_string_size_2);
      BleApp_SendUartStream(&command_list_string_3, command_list_string_size_3);

  }
  /*Check if command will be sent by can or lin*/
  else if (StrCompare(command_string, &linprotocol_string[0], 3) == 0){

    /*lin*/

     if (StrCompare(((char *)&command_string[4]), &linon_string[0], 2) == 0){
       
        LIN_SendWakeupSignal(MASTER_INSTANCE_LIN); 
        BleApp_SendUartStream(&success_string, success_string_size);
        
        if (!first_time_command_flag)
        {
           first_time_command_flag = 1;
        }
        
    } 
    else if (StrCompare(((char *)&command_string[4]), &linoff_string[0], 3) == 0){
      
        lin_go_to_sleep();
        BleApp_SendUartStream(&success_string, success_string_size);
        
        if (!first_time_command_flag)
        {
           first_time_command_flag = 1;
        }
        
    } else {
    
        if (check_if_lin_sleeping() == 1){
          
          BleApp_SendUartStream(&sleeping_string, sleeping_string_size);

        } else {

           if (StrCompare(((char *)&command_string[4]), &ledred_string[0], 14) == 0)
           {

                g_lin_frame_data_buffer[0] = gCommand_ToggleLedRed_c; 
                g_lin_frame_data_buffer[1] = ++seq_number_ledred; 
        
            } 
            else if (StrCompare(((char *)&command_string[4]), &ledgreen_string[0], 16) == 0)
            {

                g_lin_frame_data_buffer[0] = gCommand_ToggleLedGreen_c;   
                g_lin_frame_data_buffer[1] = ++seq_number_ledgreen; 
      
            }
            else if (StrCompare(((char *)&command_string[4]), &ledblue_string[0], 15) == 0)
            {

                g_lin_frame_data_buffer[0] = gCommand_ToggleLedBlue_c;    
                g_lin_frame_data_buffer[1] = ++seq_number_ledblue;  
              
            }
            else if (StrCompare(((char *)&command_string[4]), &getled_string[0], 7) == 0)
            {
       
                g_lin_frame_data_buffer[0] = gCommand_GetLed_c;
                g_lin_frame_data_buffer[1] = ++seq_number_getled; 
            }
           else if (StrCompare(((char *)&command_string[4]), &gettemp_string[0], 15) == 0)
           {
      
                g_lin_frame_data_buffer[2] = gCommand_GetTemperature_c; 
                g_lin_frame_data_buffer[3] = ++seq_number_gettemp;  
      
           } else {
                   BleApp_SendUartStream(&error_string, error_string_size);
           }

    }
    }
  }
   else if (StrCompare(command_string, &canprotocol_string[0], 3) == 0){

    /*can*/

      if (StrCompare(((char *)&command_string[4]), &ledred_string[0], 14) == 0){
        app_tx_can(gCommand_ToggleLedRed_c,0,0,0);
      }
      else if (StrCompare(((char *)&command_string[4]), &ledgreen_string[0], 16) == 0){
        app_tx_can(gCommand_ToggleLedGreen_c,0,0,0);
      }
      else if (StrCompare(((char *)&command_string[4]), &ledblue_string[0], 15) == 0){
        app_tx_can(gCommand_ToggleLedBlue_c,0,0,0);
      }
      else if (StrCompare(((char *)&command_string[4]), &gettemp_string[0], 15) == 0){
        app_tx_can(gCommand_GetTemperature_c,0,0,0);
      }
      else if (StrCompare(((char *)&command_string[4]), &getled_string[0], 7) == 0){
        app_tx_can(gCommand_GetLed_c,0,0,0);
      } else {
           BleApp_SendUartStream(&error_string, error_string_size);
    }

  }
  else{
           BleApp_SendUartStream(&error_string, error_string_size);
  }

  errorcommand_timeout_flag = 1;
  TMR_StartSingleShotTimer(mErrorCommandTimerID, 100, (pfTmrCallBack_t)mErrorCommandTimerCb, (void*)((uint32_t)mErrorCommandTimerID));

  }

}

static void mErrorCommandTimerCb
(
    uint8_t timerId
)
{
    errorcommand_timeout_flag = 0;

    timerId = timerId;  /* prevent compiler warning */
}



/*! *********************************************************************************
 * @}
 ********************************************************************************** */
